# -*- coding: utf-8 -*-
"""
Copyright (c) 2008-2024 synodriver <diguohuangjiajinweijun@gmail.com>
"""
# https://qingping.feishu.cn/docs/doccnsQEUKIl4ySLumxSqYktH4d
from dataclasses import dataclass
from enum import IntEnum
from io import BytesIO
from typing import Generator

import crcmod

crc16 = crcmod.mkCrcFun(0x13D65, 0xFFFF, True, 0xFFFF)


@dataclass
class Event:
    sop: bytes  # b"CG"  2 bytes
    cmd: int  # 1 bytes
    length: int  # 2 bytes
    payload: bytes  # len(payload) == length
    checksum: int  # 2 bytes

    @property
    def keys(self) -> dict:
        """解析payload字段"""
        ret = {}
        reader = BytesIO(self.payload)
        while True:
            key = reader.read(1)[0]
            length = int.from_bytes(reader.read(2), "little")
            value = reader.read(length)
            ret[key] = value
            if reader.tell() == len(self.payload):
                break
        return ret

    @keys.setter
    def keys(self, value) -> None:
        writer = BytesIO()
        for key, v in value.items():
            writer.write(key.to_bytes(1, "little"))
            writer.write(len(v).to_bytes(2, "little"))
            writer.write(v)
        self.payload = writer.getvalue()
        self.length = len(self.payload)

    def to_bytes(self) -> bytes:
        writer = BytesIO()
        writer.write(self.sop)
        writer.write(self.cmd.to_bytes(1, "little"))
        writer.write(self.length.to_bytes(2, "little"))
        writer.write(self.payload)
        # self.checksum = crc16(writer.getvalue())
        writer.write(self.checksum.to_bytes(2, "little"))  # todo 自己计算checksum
        return writer.getvalue()


def parse_history_data(data: bytes):
    time = int.from_bytes(data[:4], "little")  # 时间戳
    internal = int.from_bytes(data[4:6], "little")  # 存储间隔 单位s
    current_ptr = 6
    history = []
    while current_ptr != len(data):
        history.append(data[current_ptr : current_ptr + 6])
        current_ptr += 6
    return time, internal, history


def build_history_data(time: int, internal: int, history: list) -> bytes:
    writer = BytesIO()
    writer.write(time.to_bytes(4, "little"))
    writer.write(internal.to_bytes(2, "little"))
    for item in history:
        writer.write(item)
    return writer.getvalue()


class _State(IntEnum):
    read_first_5_bytes = 0
    read_payload = 1


class Connection:
    def __init__(self):
        self._buf = bytearray()  # type: bytearray
        self._state = _State.read_first_5_bytes
        self._tmp = None

    def feed_data(self, data: bytes) -> Generator[Event, None, None]:
        self._buf.extend(data)
        while True:
            if self._state == _State.read_first_5_bytes:
                if len(self._buf) < 5:
                    break
                sop: bytes = self._buf[:2]
                cmd: int = self._buf[2]
                length = int.from_bytes(self._buf[3:5], "little")
                self._tmp = Event(sop, cmd, length, b"", 0)
                del self._buf[:5]
                self._state = _State.read_payload
            elif self._state == _State.read_payload:
                if len(self._buf) < self._tmp.length + 2:
                    break
                payload = self._buf[: self._tmp.length]
                checksum = int.from_bytes(
                    self._buf[self._tmp.length : self._tmp.length + 2], "little"
                )
                del self._buf[: self._tmp.length + 2]
                self._state = _State.read_first_5_bytes
                self._tmp.payload = payload
                self._tmp.checksum = checksum
                ev = self._tmp
                self._tmp = None
                yield ev

    def send(self, event: Event) -> bytes:
        return event.to_bytes()



# -*- coding: utf-8 -*-
from typing import List, Optional, TypedDict


# {'access_token': 'xxx',
# 'expires_in': 7199,
# 'scope':  'device_full_access',
# 'token_type': 'bearer'}
class GetAccessTokenResponse(TypedDict):
    access_token: str
    refresh_token: Optional[str]
    expires_in: int
    scope: str
    token_type: str


class DeviceStatus(TypedDict):
    offline: Optional[bool]  # 状态 （True表示离线False表示在线）


class DeviceProduct(TypedDict):
    id: Optional[
        str
    ]  # 产品id https://developer.qingping.co/cloud-to-cloud/specification-guidelines#21-%E4%BA%A7%E5%93%81%E5%88%97%E8%A1%A8
    name: Optional[str]  # 产品名称
    en_name: Optional[str]  # 产品英文名称
    code: Optional[str]
    noBleSetting: Optional[bool]


class DeviceSetting(TypedDict):
    report_interval: Optional[int]  # 设备数据上报间隔（秒）
    collect_interval: Optional[int]  # 设备数据采集间隔（秒）


class DeviceInfo(TypedDict):
    name: str  # 设备名称
    mac: str  #  设备mac地址
    group_id: str  # 分组id
    group_name: str  # 分组名称
    status: Optional[DeviceStatus]  # 状态
    connection_type: Optional[str]  # WIFI
    version: str  # 设备版本
    created_at: str  # 设备注册时间
    product: DeviceProduct  # 产品信息
    setting: DeviceSetting  # 设备上报配置


class DataDetail(TypedDict):
    value: Optional[float]  # 数据值
    level: Optional[float]  # 等级
    status: Optional[float]  # 状态


class DeviceData(TypedDict):
    battery: Optional[DataDetail]  # 电量
    signal: Optional[DataDetail]  # undocumented，但是实际返回的有
    humidity: Optional[DataDetail]  # 湿度
    pressure: Optional[DataDetail]  # 气压
    temperature: Optional[DataDetail]  # 温度
    tvoc: Optional[DataDetail]  # 挥发物质
    co2: Optional[DataDetail]  # 二氧化碳
    pm25: Optional[DataDetail]  # pm2.5
    timestamp: Optional[DataDetail]  # 时间


class Device(TypedDict):
    info: DeviceInfo  # 设备信息
    data: Optional[DeviceData]  # 设备数据


class DeviceResponse(TypedDict):
    total: int  # 设备总数
    devices: List[Device]  # 设备数据


class HistoryDataResponse(TypedDict):
    total: int  # 总条数
    data: List[DeviceData]  # 数据正文


class AlertConfig(TypedDict):
    id: Optional[int]  # 配置id 仅仅在get_alert的返回值中
    metric_name: str  # 事件属性（温度、湿度、气压等）
    # （温度：temperature，湿度：humidity）
    operator: str  # 操作符（大于gt、小于lt）
    threshold: float  # 阈值


class Event(TypedDict):
    data: Optional[DeviceData]  # 设备数据
    alert_config: Optional[AlertConfig]  # 触发条件


class HistoryEventResponse(TypedDict):
    total: int  # 总条数
    events: List[Event]  # 数据正文


class GetAlertResponse(TypedDict):
    mac: str
    alert_configs: List[AlertConfig]


class Group(TypedDict):
    id: int  # 分组id
    name: str  # 分组名称


class GetGroupsResponse(TypedDict):
    total: int  # 分组总数
    groups: List[Group]  # 数据正文


class Profile(TypedDict):
    mac: str  # 设备 MAC
    ble_mac: Optional[str]  # ble.mac undocumented但是返回值有
    sn: Optional[str]  # 青萍 SN
    customization_sn: Optional[str]  # customization.sn 自定义 SN


class DeviceInfoResponse(TypedDict):
    total: int  # 总条数
    profiles: List[Profile]  # 数据正文



# -*- coding: utf-8 -*-
from unittest import TestCase

from qingping_sdk import Connection, Event, build_history_data, parse_history_data

data = bytes.fromhex(
    "43 47 31 1F 01 01 10 00 A9 F4 9F 95 FF 08 B48B 9C AB 35 8F D3 49 46 E5 02 10 00 45 30 33 30 31 35 44 37 34 39 44 45 36 32 36 35 03 F6 00 B6 88 77 5C 05 00 9A C2 2F 66 27 4E 9A C2 2F 67 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 65 27 4E 9A C2 2F 66 27 4E 9A C2 2F 66 27 4E D3 70".strip()
)
data2 = bytes.fromhex("")


class TestCon(TestCase):
    def setUp(self):
        self.con = Connection()

    def test_feed(self):
        for ev in self.con.feed_data(data):
            print(ev)
            self.assertEqual(ev.length, len(ev.payload))
            print(ev.keys)
            ev.keys = ev.keys
            his = parse_history_data(ev.keys[3])
            self.assertEqual(build_history_data(*his), ev.keys[3])

            self.assertEqual(ev.to_bytes(), data)



# -*- coding: utf-8 -*-
from qingping_sdk.client import Client
from qingping_sdk.connection import (
    Connection,
    Event,
    build_history_data,
    parse_history_data,
)

__version__ = "0.0.1"



# -*- coding: utf-8 -*-
"""
Copyright (c) 2008-2024 synodriver <diguohuangjiajinweijun@gmail.com>
"""
from base64 import b64encode


def create_auth(app_key: str, app_secret: str) -> str:
    return b64encode(app_key.encode() + b":" + app_secret.encode()).decode()



# -*- coding: utf-8 -*-
import os
import time
from unittest import IsolatedAsyncioTestCase

import dotenv

from qingping_sdk import Client

dotenv.load_dotenv("./.env")


class TestClient(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        # self.client = Client(os.getenv("APP_KEY"), os.getenv("APP_SECRET"))
        pass

    async def test_get_access_token(self):
        async with Client(os.getenv("APP_KEY"), os.getenv("APP_SECRET")) as client:
            self.assertIsNotNone(client.access_token)
            self.assertIsInstance(client.access_token, str)
            device = await client.get_devices()
            print(device)
            mac = device["devices"][0]["info"]["mac"]
            print(
                await client.get_history_data(
                    mac, int(time.time()) - 3600, int(time.time())
                )
            )
            print(
                await client.get_history_events(
                    mac, int(time.time()) - 3600, int(time.time())
                )
            )
            d = await client.change_settings([mac], 60, 60)
            self.assertEqual(d, "")

            # print(await client.add_alert(mac, {
            #     "metric_name": "temperature",
            #     "operator": "lt",
            #     "threshold": 10
            # }))
            alert = await client.get_alert(mac)
            print(alert)
            #         await client.delete_alert(mac, [403499])
            #         await client.add_alert(mac, {
            #   "metric_name": "battery",
            #   "operator": "lt",
            #   "threshold": 15
            # })
            #         alert = await client.get_alert(mac)
            # print(await client.change_alert(mac, {
            #     "id": alert["alert_configs"][0]["id"],
            #     "metric_name": "temperature",
            #     "operator": "lt",
            #     "threshold": 5
            # }))
            #
            # print(await client.delete_alert(mac, [alert["alert_configs"][0]["id"]]))
            print("---groups----")
            print(await client.get_groups())
            print("---sn----")
            print(await client.get_device_info([mac], ["sn"]))


if __name__ == "__main__":
    import unittest

    unittest.main()



# -*- coding: utf-8 -*-
"""
Copyright (c) 2008-2024 synodriver <diguohuangjiajinweijun@gmail.com>
"""
# https://developer.qingping.co/cloud-to-cloud/open-apis
import asyncio
import json
import time
from typing import List

import aiohttp

from qingping_sdk.exceptions import (
    AuthException,
    ConflictException,
    ExpiredException,
    NotFoundException,
    RequestException,
    ServerException,
)
from qingping_sdk.typing import (
    AlertConfig,
    Device,
    DeviceInfoResponse,
    DeviceResponse,
    GetAccessTokenResponse,
    GetAlertResponse,
    GetGroupsResponse,
    HistoryDataResponse,
    HistoryEventResponse,
)
from qingping_sdk.utils import create_auth

JSON_ENCODING = "utf-8"
DEFAULT_JSON_DECODER = json.loads
DEFAULT_JSON_ENCODER = json.dumps
DEFAULT_ENDPOINT = "oauth.cleargrass.com"
DEFAULT_API_ENDPOINT = "apis.cleargrass.com"


class Client:
    def __init__(
        self,
        app_key: str,
        app_secret: str,
        endpoint: str = None,
        api_endpoint: str = None,
        client_session=None,
        close_on_exit: bool = True,
        loop: asyncio.AbstractEventLoop = None,
        **kw,
    ):
        self.app_key = app_key
        self.app_secret = app_secret
        self._endpoint = endpoint or DEFAULT_ENDPOINT
        self._api_endpoint = api_endpoint or DEFAULT_API_ENDPOINT

        self.kw = kw
        self.loads = (
            self.kw.pop("loads") if "loads" in self.kw else DEFAULT_JSON_DECODER
        )  # json serialize
        self.dumps = (
            self.kw.pop("dumps") if "dumps" in self.kw else DEFAULT_JSON_ENCODER
        )
        self._loop = loop or asyncio.get_running_loop()
        self.client_session = client_session or aiohttp.ClientSession(
            json_serialize=self.dumps
        )  # aiohttp的会话
        self._close_on_exit = close_on_exit

        self._task = None
        self._ready = self._loop.create_future()
        self.access_token = None

    async def _get_access_token(self) -> GetAccessTokenResponse:
        auth_str = create_auth(self.app_key, self.app_secret)
        async with self.client_session.post(
            f"https://{self._endpoint}/oauth2/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Authorization": f"Basic {auth_str}",
            },
            data={
                "grant_type": "client_credentials",
                "scope": "device_full_access",
            },
        ) as resp:
            if resp.status != 200:
                raise AuthException(await resp.text())
            return await resp.json(loads=self.loads)

    async def aclose(self):
        await self.client_session.close()

    async def __aenter__(self):
        self._task = self._loop.create_task(self._refresh_token())
        await self._ready
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._task.cancel()
        try:
            await self._task
        except asyncio.CancelledError:
            pass
        if self._close_on_exit:
            await self.aclose()
        return False

    async def _refresh_token(self):
        while True:
            access_token_data = await self._get_access_token()
            self.access_token = access_token_data["access_token"]
            if not self._ready.done() and not self._ready.cancelled():
                self._ready.set_result(None)
            await asyncio.sleep(access_token_data["expires_in"])
        # def cb():
        #     self._loop.create_task(self._refresh_token())
        #
        # self._handle = self._loop.call_later(access_token_data["expires_in"], cb)

    async def send_request(
        self, method="GET", url: str = None, params: dict = None, json: dict = None
    ):
        url = url or f"https://{self._api_endpoint}/v1/apis/devices"
        async with self.client_session.request(
            method,
            url,
            params=params,
            json=json,
            headers={"Authorization": f"Bearer {self.access_token}"},
        ) as resp:
            if resp.status == 200:
                try:
                    return await resp.json(loads=self.loads)
                except aiohttp.client_exceptions.ContentTypeError:
                    return await resp.text()
            elif resp.status == 400:
                raise RequestException(await resp.text())
            elif resp.status in (401, 403):
                raise AuthException(await resp.text())
            elif resp.status == 404:
                raise NotFoundException(await resp.text())
            elif resp.status == 408:
                raise ExpiredException(await resp.text())
            elif resp.status == 409:
                raise ConflictException(await resp.text())
            elif resp.status in (500, 503):
                raise ServerException(await resp.text())

    async def bind_device(
        self, device_token: str, product_id: int, timestamp: int = None
    ) -> Device:
        """
        绑定设备
        :param device_token:
        :param product_id:https://developer.qingping.co/cloud-to-cloud/specification-guidelines#21-%E4%BA%A7%E5%93%81%E5%88%97%E8%A1%A8
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :return:
        """
        timestamp = timestamp or int(time.time() * 1000)
        return await self.send_request(
            "POST",
            f"https://{self._api_endpoint}/v1/apis/devices",
            json={
                "device_token": device_token,
                "product_id": product_id,
                "timestamp": timestamp,
            },
        )

    async def delete_device(self, mac: List[str], timestamp: int = None):
        timestamp = timestamp or int(time.time() * 1000)
        return await self.send_request(
            "DELETE",
            f"https://{self._api_endpoint}/v1/apis/devices",
            json={"mac": mac, "timestamp": timestamp},
        )

    async def get_devices(
        self,
        group_id: int = None,
        offset: int = None,
        limit: int = None,
        role: str = None,
        timestamp: int = None,
    ) -> DeviceResponse:
        """
        设备列表
        :param group_id:
        :param offset:
        :param limit:
        :param role: 请求者角色（只有是代理商角色需要填写这个字段，代理商标识："agent"）
        :return:
        """
        params = {"timestamp": timestamp or int(time.time() * 1000)}
        if group_id is not None:
            params["group_id"] = group_id
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit
        if role is not None:
            params["role"] = role
        return await self.send_request(
            "GET", f"https://{self._api_endpoint}/v1/apis/devices", params=params
        )

    async def get_history_data(
        self,
        mac: str,
        start_time: int,
        end_time: int,
        timestamp: int = None,
        offset: int = None,
        limit: int = None,
    ) -> HistoryDataResponse:
        """
        设备历史数据
        :param mac: 设备mac地址
        :param start_time: 开始时间戳 单位s
        :param end_time: 结束时间戳
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :param offset: 偏移量
        :param limit: 最大返回数量 不得超过200条
        :return:
        """
        params = {
            "mac": mac,
            "start_time": start_time,
            "end_time": end_time,
            "timestamp": timestamp or int(time.time() * 1000),
        }
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit
        return await self.send_request(
            "GET",
            f"https://{self._api_endpoint}/v1/apis/devices/data",
            params=params,
        )

    async def get_history_events(
        self,
        mac: str,
        start_time: int,
        end_time: int,
        timestamp: int = None,
        offset: int = None,
        limit: int = None,
    ) -> HistoryEventResponse:
        """
        设备历史事件
        :param mac: 设备mac地址
        :param start_time: 开始时间戳 单位s
        :param end_time: 结束时间戳
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :param offset: 偏移量
        :param limit: 最大返回数量 不得超过200条
        :return:
        """
        params = {
            "mac": mac,
            "start_time": start_time,
            "end_time": end_time,
            "timestamp": timestamp or int(time.time() * 1000),
        }
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit
        return await self.send_request(
            "GET",
            f"https://{self._api_endpoint}/v1/apis/devices/events",
            params=params,
        )

    async def change_settings(
        self,
        mac: List[str],
        report_interval: int,
        collect_interval: int,
        timestamp: int = None,
    ):
        """
        修改设备配置
        :param mac: 需要获取信息的设备 MAC 列表
        :param report_interval: 上报周期(秒)最小为10s,且为采集周期的整数倍
        :param collect_interval: 采集周期(秒)
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :return:
        """
        return await self.send_request(
            "PUT",
            f"https://{self._api_endpoint}/v1/apis/devices/settings",
            json={
                "mac": mac,
                "report_interval": report_interval,
                "collect_interval": collect_interval,
                "timestamp": timestamp or int(time.time() * 1000),
            },
        )

    async def add_alert(
        self, mac: str, alert_config: AlertConfig, timestamp: int = None
    ):
        return await self.send_request(
            "POST",
            f"https://{self._api_endpoint}/v1/apis/devices/settings/alert",
            json={
                "mac": mac,
                "alert_config": alert_config,
                "timestamp": timestamp or int(time.time() * 1000),
            },
        )

    async def get_alert(self, mac: str, timestamp: int = None) -> GetAlertResponse:
        return await self.send_request(
            "GET",
            f"https://{self._api_endpoint}/v1/apis/devices/settings/alert",
            params={
                "mac": mac,
                "timestamp": timestamp or int(time.time() * 1000),
            },
        )

    async def change_alert(
        self, mac: str, alert_config: AlertConfig, timestamp: int = None
    ):
        """
        修改报警配置
        :param mac:
        :param alert_config: 需要有配置id
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :return:
        """
        return await self.send_request(
            "PUT",
            f"https://{self._api_endpoint}/v1/apis/devices/settings/alert",
            json={
                "mac": mac,
                "alert_config": alert_config,
                "timestamp": timestamp or int(time.time() * 1000),
            },
        )

    async def delete_alert(self, mac: str, config_id: List[int], timestamp: int = None):
        """
        删除报警配置
        :param mac:
        :param config_id:
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :return:
        """
        return await self.send_request(
            "DELETE",
            f"https://{self._api_endpoint}/v1/apis/devices/settings/alert",
            json={
                "mac": mac,
                "config_id": config_id,
                "timestamp": timestamp or int(time.time() * 1000),
            },
        )

    async def get_groups(self, timestamp: int = None) -> GetGroupsResponse:
        """
        分组列表
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :return:
        """
        return await self.send_request(
            "GET",
            f"https://{self._api_endpoint}/v1/apis/groups",
            params={"timestamp": timestamp or int(time.time() * 1000)},
        )

    async def get_device_info(
        self, mac_list: List[str], profile: List[str], timestamp: int = None
    ) -> DeviceInfoResponse:
        """
        获取设备的基本信息，如 SN 等
        :param mac_list: 设备 MAC
        :param profile: 需要获取的设备基本信息字段列表（参考 规范说明） e.g. ["sn", "customization.sn"],
        :param timestamp: 毫秒级时间戳(13位) 20s内有效,同一个请求不可重复
        :return:
        """
        return await self.send_request(
            "POST",
            f"https://{self._api_endpoint}/v1/apis/devices/profile/query",
            json={
                "macList": mac_list,
                "profile": profile,
                "timestamp": timestamp or int(time.time() * 1000),
            },
        )



# -*- coding: utf-8 -*-


class QingpingException(Exception):
    pass


class RequestException(QingpingException):
    pass


class AuthException(QingpingException):
    pass


class NotFoundException(QingpingException):
    pass


class ConflictException(QingpingException):
    pass


class ExpiredException(QingpingException):
    pass


class ServerException(QingpingException):
    pass
